@@ -1,4 +1,4 @@
-#!/bin/sh
+#!/bin/bash
 set -e
 THIS_DIR=$(cd $(dirname $0); pwd)
 . $THIS_DIR/tools.sh
@@ -9,7 +9,7 @@ disk_size_mb="$(device_size_mb "$booted_device")"
 if [ "$((100*disk_size_mb))" -lt $((105*IMAGE_SIZE_MB)) ]
 then
     echo "Note: debootstick will not try to span partitions over this disk because it is not significantly larger than original image."
-    return
+    exit
 fi
 
 echo "** Spanning over disk space..."
